import asyncio
import atexit

from aiohttp import ClientSession, TCPConnector
from .base import INFO, SUCCESS, s_print
from .discover import Discover
from .suggested import Suggested
from .tiktoks import TikToks
from .info import Info
from .verifier import Verifier


class App:
    """
    Main application class. Initialize with app = App(proxy) to start.
    """

    _session: ClientSession = None

    @classmethod
    def __create(cls):
        if cls._session is None:
            cls._session = ClientSession(connector=TCPConnector(verify_ssl=False))

    def __init__(self, proxy: str = None):
        """
        Args:
            proxy (str): proxy server url
        """
        App.__create()
        s_print("HTTP session connected", INFO)

        self._proxy = ""
        if proxy:
            s_print(f"Connecting to proxy at {proxy}", INFO)
            self._proxy = proxy

        self.discover = Discover(self._session, self._proxy)
        self.suggested = Suggested(self._session, self._proxy)
        verifier_object = Verifier(self._session, self._proxy)
        self.tiktoks = TikToks(verifier_object)
        self.info = Info(verifier_object)

        s_print("Started component apps", INFO)

        atexit.register(App.cleanup)

        s_print("Initialized app!", SUCCESS)

    @property
    def proxy(self):
        return self._proxy

    @proxy.setter
    def proxy(self, proxy: str):
        self._proxy = proxy

    @classmethod
    def cleanup(cls):
        s_print("Closing HTTP session...", INFO)
        if cls._session:
            asyncio.get_event_loop().create_task(cls._session.close())
        s_print("App shut down.", SUCCESS)



from collections import deque
from random import sample
from typing import List
from tqdm import tqdm

import jmespath as jp
from aiohttp import ClientSession
from colored import stylize

from .base import Base, ERROR, s_print, INFO


class Suggested(Base):
    """
    Class to get suggested users, musics, and challenges.
    """

    _headers = {
        "accept": "application/json, text/plain, */*",
        "user-agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.116 Safari/537.36",
        "accept-language": "en-US,en;q=0.9",
        "accept-encoding": "gzip",
    }

    _params = {
        "noUser": 0,
        "userCount": 30,
        "userId": "",
    }

    _url = "https://m.tiktok.com/node/share/discover"

    def __init__(self, session: ClientSession, proxy: str):
        super().__init__(session, proxy)

    @staticmethod
    def __separate_results(results_list: list, no_user: bool) -> dict:
        """
        Transform the results of finding suggested users, musics, and challenges from list of lists to dict of lists

        Args:
            results_list (list): list of lists (each child list corresponds to either user, challege, and music
            no_user (bool): whether there is a list of users at the start of the list or not

        Returns:
            dict: converted dictionary

        """
        if no_user:
            return {"challenge": results_list[0], "music": results_list[1]}
        return {
            "user": results_list[0],
            "challenge": results_list[1],
            "music": results_list[2],
        }

    @classmethod
    def __merge_crawl_results(cls, results_list: List[dict]) -> dict:
        """
        Merge together the results of crawling from a list of many dicts sharing
        the same keys to one dict with those keys

        Args:
            results_list (list): list of dicts with 'users', 'challenges', and 'musics' keys

        Returns:
            dict: dict merged from the lists
        """
        users = challenges = musics = []
        for result_dict in results_list:
            users.extend(result_dict.get("user", []))
            challenges.extend(result_dict.get("challenge", []))
            musics.extend(result_dict.get("music", []))
        return cls.__separate_results([users, challenges, musics], False)

    async def fetch(
        self, user_id: str = None, user_count: int = 30, no_user: bool = False, **kwargs
    ) -> dict:
        """
        Fetch suggested users, musics, and challenges.

        Args:
            user_id (int): ID of the TikTok user for whom suggestions are made
            user_count (int): number of suggested users (only) to return from the call
            no_user (bool): whether or not to return suggested users
            **kwargs: any other path parameters

        Returns:
            dict: dict of the suggested elements
        """
        if user_count > 99:
            raise ValueError(("user_count must be < 100", "red"))

        url = self._url

        explicit_kwargs = {
            "user_id": user_id,
            "user_count": user_count,
            "no_user": int(no_user),
        }

        response = await self._get_data(
            url, self._headers, explicit_kwargs, self._proxy, **kwargs
        )

        return self.__separate_results(
            jp.search("body[*].exploreList", response), no_user
        )

    async def crawl(
        self,
        depth: int = 1,
        choice_count: int = 1,
        user_id: str = None,
        user_count: int = 30,
        **kwargs
    ) -> dict:
        """
        Crawl many suggested elements by using a spider

        Args:
            depth (int): the depth of crawling (depth * choice_count = number of calls made)
            choice_count (int): how many user IDs to sample from each result set to crawl on next
            user_id (int): starting user ID
            user_count (int): how many suggested users to return from each call
            **kwargs: other path parameters

        Returns:
            dict: dict of crawled suggested elements
        """
        if user_count < choice_count or user_count < 10:
            raise ValueError(
                stylize("user_count must be >= choice_count and >= 10", ERROR)
            )

        user_ids_queue = deque([user_id])
        limit = depth * choice_count
        level = 0
        results = []

        s_print("Starting crawler...", INFO)

        with tqdm(total=(choice_count ** 2) * depth + 1) as pbar:
            while len(user_ids_queue):
                next_user_id = user_ids_queue.popleft()

                items = await self.fetch(next_user_id, user_count, False, **kwargs)
                results.append(items)

                response_user_ids = [
                    item["cardItem"]["id"] for item in items.get("user", [])
                ]

                pbar.update()
                if level < limit:
                    user_ids_queue.extend(sample(response_user_ids, choice_count))
                    level += 1

        return self.__merge_crawl_results(results)



import colored
import orjson
from aiohttp import ClientSession
from colored import stylize
from humps import decamelize
from inflection import camelize


SUCCESS = colored.fg("light_green") + colored.attr("bold")
INFO = colored.fg("aquamarine_3")
ERROR = colored.fg("red")


def s_print(text: str, style: colored):
    """
    Print text with style in the terminal

    Args:
        text (str): text to print
        style (colored): colored style to print with

    Returns:
        None: None
    """
    print(stylize(text, style))


class Base:
    """
    Base class used to construct other classes.
    """
    _headers: dict = {}
    _params: dict = {}
    _proxy: str = ""
    _url: str = ""
    _urls: list = []
    _session: ClientSession = None

    @classmethod
    def init_class(cls, session: ClientSession, proxy: str):
        if cls._session is None:
            cls._session = session
        if not cls._proxy:
            cls._proxy = proxy

    def __init__(self, session: ClientSession, proxy: str):
        self.init_class(session, proxy)

    @classmethod
    def get_default_params(cls):
        """
        Returns:
            dict: the default parameters to be used in music, user, and challenge

        """
        return decamelize(cls._params)

    @classmethod
    def __convert_options(cls, options: dict = None) -> dict:
        """
        Call .utils.options_to_params to convert keyword options to params

        Args:
            options (dict): keyword options to convert

        Returns:
            dict: dict of converted path parameters

        """
        new_options = {**cls._params, **options}
        return cls.__options_to_params(new_options)

    @staticmethod
    def __options_to_params(options: dict) -> dict:
        """
        Convert keywords options to a dict of RESTful parameters

        Args:
            options (dict): dict of keyword options

        Returns:
            dict: dict of parameterized options

        """
        params = {}
        for key, val in options.items():
            camel_key = camelize(key, uppercase_first_letter=False)
            params[camel_key] = "" if val is None else str(val)
        return params

    async def _get_data(
        self, url: str, headers, params: dict = None, proxy: str = None, **kwargs
    ) -> dict:
        """
        Async get URL and return a dict of results read using ORJSON and aiohttp

        Args:
            url (str): url to get
            headers (dict): dict of headers
            params (dict): dict of path parameters
            proxy (str): http url of proxy server
            **kwargs:

        Returns:
            dict: response converted to dict

        """
        converted_params = self.__convert_options({**params, **kwargs})
        async with self._session.get(
            url, headers=headers, params=converted_params, proxy=proxy
        ) as response:
            content = await response.content.read()
        return orjson.loads(content)



from .verifier import Verifier
from .base import Base

# TODO: implement page scrolling


class TikToks(Base):
    """
    Class to fetch tiktoks owned by an user, music, or challenge
    """

    _url = "https://m.tiktok.com/share/item/list?verifyFp="
    _params = {
        "secUid": "",
        "id": "",
        "type": "",
        "count": 30,
        "minCursor": 0,
        "maxCursor": 0,
        "shareUid": "",
        "lang": "en",
        "verifyFp": "",
    }

    _headers = {
        "authority": "m.tiktok.com",
        "method": "GET",
        "scheme": "https",
        "accept": "application/json, text/plain, */*",
        "accept-encoding": "gzip, deflate, br",
        "accept-language": "en-US,en;q=0.9",
        "referrer": "https://www.tiktok.com/",
        "sec-fetch-dest": "empty",
        "sec-fetch-mode": "cors",
        "sec-fetch-site": "same-site",
        "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.0 Safari/537.36)",
        "path": "",
    }

    def __init__(self, verifier: Verifier):
        self._verifier = verifier
        super().__init__(verifier.session, verifier.proxy)

    async def __prep(self, url):
        """
        Sign the url with fingerprint

        Args:
            url (str): url to sign

        Returns:
            tuple: signed headers & parameters

        """
        headers = self._headers.copy()
        headers["path"] = url.split("tiktok.com")[1]

        verifiers = await self._verifier.get_verifiers(url)
        explicit_kwargs = self._params.copy()
        explicit_kwargs.update(verifiers)
        return headers, explicit_kwargs

    async def __fetch_and_scroll(
        self,
        total: int,
        url: str,
        headers: dict,
        explicit_kwargs: dict,
        proxy: str,
        **kwargs
    ):
        """
        Get any number of tiktoks by implementing scrolling cursors
        Args:
            total (int): total number of tiktoks wanted
            url (str): url
            headers (dict): headers
            explicit_kwargs (dict): parameters
            proxy (str): proxy
            **kwargs (dict): any other path parameters

        Returns:
            list: list of tiktoks fetched

        """
        results: list = []
        max_cursor: int = 0
        has_more = True
        count = total

        while len(results) < total and has_more:
            explicit_kwargs = {
                **explicit_kwargs,
                **dict(count=count, minCursor=0, maxCursor=max_cursor),
            }
            r = await self._get_data(url, headers, explicit_kwargs, proxy, **kwargs)
            results.extend(r["body"]["itemListData"])
            max_cursor = r["body"]["maxCursor"]
            count = total - len(results)
            has_more = r["body"]["hasMore"]

        return results

    async def from_music_id(self, music_id: int, total: int = 30, **kwargs):
        """
        Get tiktoks owned by a music

        Args:
            music_id (int): numeric id of the music
            total (int): total number of tiktoks wanted
            **kwargs: any other path paramters

        Returns:
            list: list of tiktoks owned by the music

        """
        url = self._url

        headers, explicit_kwargs = await self.__prep(url)

        explicit_kwargs["type"] = 4
        explicit_kwargs["id"] = music_id

        return await self.__fetch_and_scroll(
            total, url, headers, explicit_kwargs, self._proxy, **kwargs
        )

    async def from_user_id(self, user_id: int, total: int = 30, **kwargs):
        """
        Get tiktoks owned by a user

        Args:
            user_id (int): numeric id of the user
            total (int): total number of tiktoks wanted
            **kwargs: any other path paramters

        Returns:
            list: list of tiktoks owned by the user

        """
        url = self._url

        headers, explicit_kwargs = await self.__prep(url)

        explicit_kwargs["type"] = 1
        explicit_kwargs["id"] = user_id

        return await self.__fetch_and_scroll(
            total, url, headers, explicit_kwargs, self._proxy, **kwargs
        )

    async def from_challenge_id(self, challenge_id: int, total: int = 30, **kwargs):
        """
        Get tiktoks owned by a challenge

        Args:
            challenge_id (int): numeric id of the hashtag challenge
            total (int): total number of tiktoks wanted
            **kwargs: any other path paramters

        Returns:
            list: list of tiktoks owned by the challenge

        """
        url = self._url

        headers, explicit_kwargs = await self.__prep(url)

        explicit_kwargs["type"] = 3
        explicit_kwargs["id"] = challenge_id

        return await self.__fetch_and_scroll(
            total, url, headers, explicit_kwargs, self._proxy, **kwargs
        )



import asyncio
import random
import string
from contextlib import suppress

from aiohttp import ClientSession
from pyppeteer import launch


class Verifier:
    """
    Class for verifying TikTok fingerprint
    """

    browser = None
    page = None
    redirect_url: str

    def __init__(
        self, session: ClientSession, proxy: str = "", find_redirect: bool = False
    ):
        self._session = session
        self._proxy = proxy
        self.user_agent = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.0 Safari/537.36)"
        self.args = [
            "--no-sandbox",
            "--disable-setuid-sandbox",
            "--disable-infobars",
            "--window-position=0,0",
            "--ignore-certifcate-errors",
            "--ignore-certifcate-errors-spki-list",
            "--user-agent=" + self.user_agent,
        ]

        if proxy and "@" in proxy:
            self.args.append(
                "--proxy-server="
                + proxy.split(":")[0]
                + "://"
                + proxy.split("://")[1].split(":")[1].split("@")[1]
                + ":"
                + proxy.split("://")[1].split(":")[2]
            )
        else:
            self.args.append("--proxy-server=" + proxy)

        self.options = {
            "args": self.args,
            "headless": True,
            "ignoreHTTPSErrors": True,
            "userDataDir": "./tmp",
            "handleSIGINT": False,
            "handleSIGTERM": False,
            "handleSIGHUP": False,
        }

        self.find_redirect = find_redirect

        asyncio.get_event_loop().create_task(self.start_up())

    @property
    def session(self):
        return self._session

    @property
    def proxy(self):
        return self._proxy

    async def start_up(self):
        self.browser = await launch(self.options)
        self.page = await self.browser.newPage()

        await self.page.evaluateOnNewDocument(
            """() => {
             delete navigator.__proto__.webdriver;
         }"""
        )

        # Check for user:pass proxy
        if self._proxy and "@" in self._proxy:
            await self.page.authenticate(
                {
                    "username": self._proxy.split("://")[1].split(":")[0],
                    "password": self._proxy.split("://")[1].split(":")[1].split("@")[0],
                }
            )

    async def get_verifiers(self, url: str) -> dict:
        if self.find_redirect:
            await self.go_find_redirect(url)

        # might have to switch to a tiktok url if they improve security
        await self.page.goto("about:blank", {"waitUntil": "load"})

        self.user_agent = await self.page.evaluate(
            """() => {return navigator.userAgent; }"""
        )

        verify_fp = "".join(
            random.choice(
                string.ascii_lowercase + string.ascii_uppercase + string.digits
            )
            for _ in range(16)
        )

        await self.page.evaluate("() => { " + await self.__get_js() + " }")

        signature = await self.page.evaluate(
            '''() => {
            var url = "'''
            + url
            + "&verifyFp="
            + verify_fp
            + """"
            var token = window.byted_acrawler.sign({url: url});
            return token;
        }"""
        )

        return {"verifyFp": verify_fp, "_signature": signature}

    async def go_find_redirect(self, url: str):
        with suppress(Exception):
            await self.page.goto(url, {"waitUntil": "load"})

            self.redirect_url = self.page.url

        await self.browser.close()

    async def __get_js(self):
        r = await self._session.get(
            "https://sf16-muse-va.ibytedtos.com/obj/rc-web-sdk-gcs/acrawler.js",
            proxy=self._proxy,
        )
        return await r.text()



from aiohttp import ClientSession
from .base import Base


# TODO: implement offsets


class Discover(Base):
    """
    Class to get users, musics, and challenges from the Discover page.
    """

    _headers = {
        "accept": "application/json, text/plain, */*",
        "accept-encoding": "gzip",
        "accept-language": "en-US,en;q=0.9",
    }

    _params = {
        "discoverType": 0,
        "needItemList": True,
        "keyWord": "",
        "offset": 0,
        "count": 30,
        "useRecommend": False,
        "language": "en",
    }

    _urls = {
        "music": "https://m.tiktok.com/api/discover/music/",
        "user": "https://m.tiktok.com/api/discover/user/",
        "challenge": "https://m.tiktok.com/api/discover/challenge/",
    }

    def __init__(self, session: ClientSession, proxy: str):
        super().__init__(session, proxy)

    async def music(self, **kwargs):
        """
        Return musics on the Discover page

        Args:
            **kwargs: any path_parameters
        Notes:
            Experiment with parameters [need_item_list, key_word, offset, count, use_recommend, language]
        Returns:
            dict: musics from the Discover page

        """
        url = self._urls.get("music")
        return await self._get_data(url, self._headers, kwargs, self._proxy)

    async def user(self, **kwargs):
        """
        Return users on the Discover page

        Args:
            **kwargs: any path_parameters
        Notes:
            Experiment with parameters [need_item_list, key_word, offset, count, use_recommend, language]
        Returns:
            dict: users from the Discover page

        """
        url = self._urls.get("user")
        return await self._get_data(url, self._headers, kwargs, self._proxy)

    async def challenge(self, **kwargs):
        """
        Return challenges on the Discover page

        Args:
            **kwargs: any path_parameters
        Notes:
            Experiment with parameters [need_item_list, key_word, offset, count, use_recommend, language]
        Returns:
            dict: challenges from the Discover page

        """
        url = self._urls.get("challenge")
        return await self._get_data(url, self._headers, kwargs, self._proxy)



from .verifier import Verifier
from .base import Base
from requests.models import PreparedRequest


class Info(Base):
    """
    Class for getting information about a single user, music, or challenge
    """

    _urls = {
        "music": "https://m.tiktok.com/api/music/detail/",
        "user": "https://m.tiktok.com/api/user/detail/",
        "challenge": "https://m.tiktok.com/api/challenge/detail/",
    }

    _params = {"language": "en"}

    _headers = {
        "authority": "m.tiktok.com",
        "method": "GET",
        "scheme": "https",
        "accept": "application/json, text/plain, */*",
        "accept-encoding": "gzip, deflate, br",
        "accept-language": "en-US,en;q=0.9",
        "referrer": "https://www.tiktok.com/",
        "sec-fetch-dest": "empty",
        "sec-fetch-mode": "cors",
        "sec-fetch-site": "same-site",
        "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.0 Safari/537.36)",
        "path": "",
    }

    def __init__(self, verifier: Verifier):
        self._verifier = verifier
        super().__init__(verifier.session, verifier.proxy)

    async def __mid_prep(self, url: str, explicit_kwargs: dict):
        """
        Prep headers and parameters by signing the url and adding headers

        Args:
            url (str): url to sign
            explicit_kwargs (str): parameters to add for signing

        Returns:
            tuple: signed url & signed headers

        """
        req = PreparedRequest()
        req.prepare_url(url, explicit_kwargs)
        req.prepare_url(req.url, await self._verifier.get_verifiers(req.url))
        headers = self._headers.copy()
        headers["path"] = req.url.split("tiktok.com")[1]
        return req.url, headers

    async def music(self, music_id: int, **kwargs):
        """
        Get music info by id

        Args:
            music_id (int): numeric id of music to get
            **kwargs: any other path parameters

        Returns:
            dict: music info

        """
        explicit_kwargs = self._params.copy()
        explicit_kwargs["musicId"] = str(music_id)

        url, headers = await self.__mid_prep(self._urls["music"], explicit_kwargs)

        return await self._get_data(url, headers, kwargs, self._proxy)

    async def challenge_by_id(self, challenge_id: int, **kwargs):
        """
        Get challenge info by id

        Args:
            challenge_id (int): numeric id of challenge to get
            **kwargs: any other path parameters

        Returns:
            dict: challenge info

        """
        explicit_kwargs = self._params.copy()
        explicit_kwargs["challengeId"] = str(challenge_id)

        url, headers = await self.__mid_prep(self._urls["challenge"], explicit_kwargs)

        return await self._get_data(url, headers, kwargs, self._proxy)

    async def challenge_by_name(self, challenge_name: str, **kwargs):
        """
        Get challenge info by id

        Args:
            challenge_name (str): name of the hastag challenge to get (without #)
            **kwargs: any other path parameters

        Returns:
            dict: challenge info

        """
        explicit_kwargs = self._params.copy()
        explicit_kwargs["challengeName"] = challenge_name

        url, headers = await self.__mid_prep(self._urls["challenge"], explicit_kwargs)

        return await self._get_data(url, headers, kwargs, self._proxy)

    async def user_by_name(self, user_name: str, **kwargs):
        """
        Get user info by id

        Args:
            user_name (str): name of the user to get (unique name)
            **kwargs: any other path parameters

        Returns:
            dict: user info

        """
        explicit_kwargs = self._params.copy()
        explicit_kwargs["uniqueId"] = user_name

        url, headers = await self.__mid_prep(self._urls["user"], explicit_kwargs)

        return await self._get_data(url, headers, kwargs, self._proxy)

    async def user_by_id(self, user_id: int, **kwargs):
        """
        Get user info by id

        Args:
            user_id (int): numeric id of user to get
            **kwargs: any other path parameters

        Returns:
            dict: user info

        """
        explicit_kwargs = self._params.copy()
        explicit_kwargs["userId"] = str(user_id)

        url, headers = await self.__mid_prep(self._urls["user"], explicit_kwargs)

        return await self._get_data(url, headers, kwargs, self._proxy)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="piktok",
    version="0.1.5",
    author="Anh Chu",
    author_email="chuducanh.atred@gmail.com",
    description="A fast module to retrieve data from TikTok",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/anh-chu/piktok",
    packages=setuptools.find_packages(),
    install_requires=[
        "requests",
        "orjson",
        "aiohttp",
        "pyhumps",
        "inflection",
        "jmespath",
        "pyppeteer",
        "colored",
        "tqdm",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)
